# Copyright (C) 2010 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""An OAuth 2.0 client.

Tools for interacting with OAuth 2.0 protected resources.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'

import base64
from . import clientsecrets
import copy
import datetime
import httplib2
import logging
import os
import sys
import time
import urllib.request, urllib.parse, urllib.error
import urllib.parse

from medialib.youtube.oauth2client import GOOGLE_AUTH_URI
from medialib.youtube.oauth2client import GOOGLE_REVOKE_URI
from medialib.youtube.oauth2client import GOOGLE_TOKEN_URI
from medialib.youtube.oauth2client import util
from medialib.youtube.oauth2client.anyjson import simplejson

HAS_OPENSSL = False
HAS_CRYPTO = False
try:
  from oauth2client import crypt
  HAS_CRYPTO = True
  if crypt.OpenSSLVerifier is not None:
    HAS_OPENSSL = True
except ImportError:
  pass

try:
  from urllib.parse import parse_qsl
except ImportError:
  from cgi import parse_qsl

logger = logging.getLogger(__name__)

# Expiry is stored in RFC3339 UTC format
EXPIRY_FORMAT = '%Y-%m-%dT%H:%M:%SZ'

# Which certs to use to validate id_tokens received.
ID_TOKEN_VERIFICATON_CERTS = 'https://www.googleapis.com/oauth2/v1/certs'

# Constant to use for the out of band OAuth 2.0 flow.
OOB_CALLBACK_URN = 'urn:ietf:wg:oauth:2.0:oob'

# Google Data client libraries may need to set this to [401, 403].
REFRESH_STATUS_CODES = [401]


class Error(Exception):
  """Base error for this module."""


class FlowExchangeError(Error):
  """Error trying to exchange an authorization grant for an access token."""


class AccessTokenRefreshError(Error):
  """Error trying to refresh an expired access token."""


class TokenRevokeError(Error):
  """Error trying to revoke a token."""


class UnknownClientSecretsFlowError(Error):
  """The client secrets file called for an unknown type of OAuth 2.0 flow. """


class AccessTokenCredentialsError(Error):
  """Having only the access_token means no refresh is possible."""


class VerifyJwtTokenError(Error):
  """Could on retrieve certificates for validation."""


class NonAsciiHeaderError(Error):
  """Header names and values must be ASCII strings."""


def _abstract():
  raise NotImplementedError('You need to override this function')


class MemoryCache(object):
  """httplib2 Cache implementation which only caches locally."""

  def __init__(self):
    self.cache = {}

  def get(self, key):
    return self.cache.get(key)

  def set(self, key, value):
    self.cache[key] = value

  def delete(self, key):
    self.cache.pop(key, None)


class Credentials(object):
  """Base class for all Credentials objects.

  Subclasses must define an authorize() method that applies the credentials to
  an HTTP transport.

  Subclasses must also specify a classmethod named 'from_json' that takes a JSON
  string as input and returns an instaniated Credentials object.
  """

  NON_SERIALIZED_MEMBERS = ['store']

  def authorize(self, http):
    """Take an httplib2.Http instance (or equivalent) and authorizes it.

    Authorizes it for the set of credentials, usually by replacing
    http.request() with a method that adds in the appropriate headers and then
    delegates to the original Http.request() method.

    Args:
      http: httplib2.Http, an http object to be used to make the refresh
        request.
    """
    _abstract()

  def refresh(self, http):
    """Forces a refresh of the access_token.

    Args:
      http: httplib2.Http, an http object to be used to make the refresh
        request.
    """
    _abstract()

  def revoke(self, http):
    """Revokes a refresh_token and makes the credentials void.

    Args:
      http: httplib2.Http, an http object to be used to make the revoke
        request.
    """
    _abstract()

  def apply(self, headers):
    """Add the authorization to the headers.

    Args:
      headers: dict, the headers to add the Authorization header to.
    """
    _abstract()

  def _to_json(self, strip):
    """Utility function that creates JSON repr. of a Credentials object.

    Args:
      strip: array, An array of names of members to not include in the JSON.

    Returns:
       string, a JSON representation of this instance, suitable to pass to
       from_json().
    """
    t = type(self)
    d = copy.copy(self.__dict__)
    for member in strip:
      if member in d:
        del d[member]
    if 'token_expiry' in d and isinstance(d['token_expiry'], datetime.datetime):
      d['token_expiry'] = d['token_expiry'].strftime(EXPIRY_FORMAT)
    # Add in information we will need later to reconsistitue this instance.
    d['_class'] = t.__name__
    d['_module'] = t.__module__
    return simplejson.dumps(d)

  def to_json(self):
    """Creating a JSON representation of an instance of Credentials.

    Returns:
       string, a JSON representation of this instance, suitable to pass to
       from_json().
    """
    return self._to_json(Credentials.NON_SERIALIZED_MEMBERS)

  @classmethod
  def new_from_json(cls, s):
    """Utility class method to instantiate a Credentials subclass from a JSON
    representation produced by to_json().

    Args:
      s: string, JSON from to_json().

    Returns:
      An instance of the subclass of Credentials that was serialized with
      to_json().
    """
    data = simplejson.loads(s)
    # Find and call the right classmethod from_json() to restore the object.
    module = data['_module']
    try:
      m = __import__(module)
    except ImportError:
      # In case there's an object from the old package structure, update it
      module = module.replace('.apiclient', '')
      m = __import__(module)

    m = __import__(module, fromlist=module.split('.')[:-1])
    kls = getattr(m, data['_class'])
    from_json = getattr(kls, 'from_json')
    return from_json(s)

  @classmethod
  def from_json(cls, s):
    """Instantiate a Credentials object from a JSON description of it.

    The JSON should have been produced by calling .to_json() on the object.

    Args:
      data: dict, A deserialized JSON object.

    Returns:
      An instance of a Credentials subclass.
    """
    return Credentials()


class Flow(object):
  """Base class for all Flow objects."""
  pass


class Storage(object):
  """Base class for all Storage objects.

  Store and retrieve a single credential. This class supports locking
  such that multiple processes and threads can operate on a single
  store.
  """

  def acquire_lock(self):
    """Acquires any lock necessary to access this Storage.

    This lock is not reentrant.
    """
    pass

  def release_lock(self):
    """Release the Storage lock.

    Trying to release a lock that isn't held will result in a
    RuntimeError.
    """
    pass

  def locked_get(self):
    """Retrieve credential.

    The Storage lock must be held when this is called.

    Returns:
      oauth2client.client.Credentials
    """
    _abstract()

  def locked_put(self, credentials):
    """Write a credential.

    The Storage lock must be held when this is called.

    Args:
      credentials: Credentials, the credentials to store.
    """
    _abstract()

  def locked_delete(self):
    """Delete a credential.

    The Storage lock must be held when this is called.
    """
    _abstract()

  def get(self):
    """Retrieve credential.

    The Storage lock must *not* be held when this is called.

    Returns:
      oauth2client.client.Credentials
    """
    self.acquire_lock()
    try:
      return self.locked_get()
    finally:
      self.release_lock()

  def put(self, credentials):
    """Write a credential.

    The Storage lock must be held when this is called.

    Args:
      credentials: Credentials, the credentials to store.
    """
    self.acquire_lock()
    try:
      self.locked_put(credentials)
    finally:
      self.release_lock()

  def delete(self):
    """Delete credential.

    Frees any resources associated with storing the credential.
    The Storage lock must *not* be held when this is called.

    Returns:
      None
    """
    self.acquire_lock()
    try:
      return self.locked_delete()
    finally:
      self.release_lock()


def clean_headers(headers):
  """Forces header keys and values to be strings, i.e not unicode.

  The httplib module just concats the header keys and values in a way that may
  make the message header a unicode string, which, if it then tries to
  contatenate to a binary request body may result in a unicode decode error.

  Args:
    headers: dict, A dictionary of headers.

  Returns:
    The same dictionary but with all the keys converted to strings.
  """
  clean = {}
  try:
    for k, v in list(headers.items()):
      clean[str(k)] = str(v)
  except UnicodeEncodeError:
    raise NonAsciiHeaderError(k + ': ' + v)
  return clean


def _update_query_params(uri, params):
  """Updates a URI with new query parameters.

  Args:
    uri: string, A valid URI, with potential existing query parameters.
    params: dict, A dictionary of query parameters.

  Returns:
    The same URI but with the new query parameters added.
  """
  parts = list(urllib.parse.urlparse(uri))
  query_params = dict(parse_qsl(parts[4])) # 4 is the index of the query part
  query_params.update(params)
  parts[4] = urllib.parse.urlencode(query_params)
  return urllib.parse.urlunparse(parts)


class OAuth2Credentials(Credentials):
  """Credentials object for OAuth 2.0.

  Credentials can be applied to an httplib2.Http object using the authorize()
  method, which then adds the OAuth 2.0 access token to each request.

  OAuth2Credentials objects may be safely pickled and unpickled.
  """

  @util.positional(8)
  def __init__(self, access_token, client_id, client_secret, refresh_token,
               token_expiry, token_uri, user_agent, revoke_uri=None,
               id_token=None, token_response=None):
    """Create an instance of OAuth2Credentials.

    This constructor is not usually called by the user, instead
    OAuth2Credentials objects are instantiated by the OAuth2WebServerFlow.

    Args:
      access_token: string, access token.
      client_id: string, client identifier.
      client_secret: string, client secret.
      refresh_token: string, refresh token.
      token_expiry: datetime, when the access_token expires.
      token_uri: string, URI of token endpoint.
      user_agent: string, The HTTP User-Agent to provide for this application.
      revoke_uri: string, URI for revoke endpoint. Defaults to None; a token
        can't be revoked if this is None.
      id_token: object, The identity of the resource owner.
      token_response: dict, the decoded response to the token request. None
        if a token hasn't been requested yet. Stored because some providers
        (e.g. wordpress.com) include extra fields that clients may want.

    Notes:
      store: callable, A callable that when passed a Credential
        will store the credential back to where it came from.
        This is needed to store the latest access_token if it
        has expired and been refreshed.
    """
    self.access_token = access_token
    self.client_id = client_id
    self.client_secret = client_secret
    self.refresh_token = refresh_token
    self.store = None
    self.token_expiry = token_expiry
    self.token_uri = token_uri
    self.user_agent = user_agent
    self.revoke_uri = revoke_uri
    self.id_token = id_token
    self.token_response = token_response

    # True if the credentials have been revoked or expired and can't be
    # refreshed.
    self.invalid = False

  def authorize(self, http):
    """Authorize an httplib2.Http instance with these credentials.

    The modified http.request method will add authentication headers to each
    request and will refresh access_tokens when a 401 is received on a
    request. In addition the http.request method has a credentials property,
    http.request.credentials, which is the Credentials object that authorized
    it.

    Args:
       http: An instance of httplib2.Http
         or something that acts like it.

    Returns:
       A modified instance of http that was passed in.

    Example:

      h = httplib2.Http()
      h = credentials.authorize(h)

    You can't create a new OAuth subclass of httplib2.Authenication
    because it never gets passed the absolute URI, which is needed for
    signing. So instead we have to overload 'request' with a closure
    that adds in the Authorization header and then calls the original
    version of 'request()'.
    """
    request_orig = http.request

    # The closure that will replace 'httplib2.Http.request'.
    @util.positional(1)
    def new_request(uri, method='GET', body=None, headers=None,
                    redirections=httplib2.DEFAULT_MAX_REDIRECTS,
                    connection_type=None):
      if not self.access_token:
        logger.info('Attempting refresh to obtain initial access_token')
        self._refresh(request_orig)

      # Modify the request headers to add the appropriate
      # Authorization header.
      if headers is None:
        headers = {}
      self.apply(headers)

      if self.user_agent is not None:
        if 'user-agent' in headers:
          headers['user-agent'] = self.user_agent + ' ' + headers['user-agent']
        else:
          headers['user-agent'] = self.user_agent

      resp, content = request_orig(uri, method, body, clean_headers(headers),
                                   redirections, connection_type)

      if resp.status in REFRESH_STATUS_CODES:
        logger.info('Refreshing due to a %s' % str(resp.status))
        self._refresh(request_orig)
        self.apply(headers)
        return request_orig(uri, method, body, clean_headers(headers),
                            redirections, connection_type)
      else:
        return (resp, content)

    # Replace the request method with our own closure.
    http.request = new_request

    # Set credentials as a property of the request method.
    setattr(http.request, 'credentials', self)

    return http

  def refresh(self, http):
    """Forces a refresh of the access_token.

    Args:
      http: httplib2.Http, an http object to be used to make the refresh
        request.
    """
    self._refresh(http.request)

  def revoke(self, http):
    """Revokes a refresh_token and makes the credentials void.

    Args:
      http: httplib2.Http, an http object to be used to make the revoke
        request.
    """
    self._revoke(http.request)

  def apply(self, headers):
    """Add the authorization to the headers.

    Args:
      headers: dict, the headers to add the Authorization header to.
    """
    headers['Authorization'] = 'Bearer ' + self.access_token

  def to_json(self):
    return self._to_json(Credentials.NON_SERIALIZED_MEMBERS)

  @classmethod
  def from_json(cls, s):
    """Instantiate a Credentials object from a JSON description of it. The JSON
    should have been produced by calling .to_json() on the object.

    Args:
      data: dict, A deserialized JSON object.

    Returns:
      An instance of a Credentials subclass.
    """
    data = simplejson.loads(s)
    if 'token_expiry' in data and not isinstance(data['token_expiry'],
        datetime.datetime):
      try:
        data['token_expiry'] = datetime.datetime.strptime(
            data['token_expiry'], EXPIRY_FORMAT)
      except:
        data['token_expiry'] = None
    retval = cls(
        data['access_token'],
        data['client_id'],
        data['client_secret'],
        data['refresh_token'],
        data['token_expiry'],
        data['token_uri'],
        data['user_agent'],
        revoke_uri=data.get('revoke_uri', None),
        id_token=data.get('id_token', None),
        token_response=data.get('token_response', None))
    retval.invalid = data['invalid']
    return retval

  @property
  def access_token_expired(self):
    """True if the credential is expired or invalid.

    If the token_expiry isn't set, we assume the token doesn't expire.
    """
    if self.invalid:
      return True

    if not self.token_expiry:
      return False

    now = datetime.datetime.utcnow()
    if now >= self.token_expiry:
      logger.info('access_token is expired. Now: %s, token_expiry: %s',
                  now, self.token_expiry)
      return True
    return False

  def set_store(self, store):
    """Set the Storage for the credential.

    Args:
      store: Storage, an implementation of Stroage object.
        This is needed to store the latest access_token if it
        has expired and been refreshed. This implementation uses
        locking to check for updates before updating the
        access_token.
    """
    self.store = store

  def _updateFromCredential(self, other):
    """Update this Credential from another instance."""
    self.__dict__.update(other.__getstate__())

  def __getstate__(self):
    """Trim the state down to something that can be pickled."""
    d = copy.copy(self.__dict__)
    del d['store']
    return d

  def __setstate__(self, state):
    """Reconstitute the state of the object from being pickled."""
    self.__dict__.update(state)
    self.store = None

  def _generate_refresh_request_body(self):
    """Generate the body that will be used in the refresh request."""
    body = urllib.parse.urlencode({
        'grant_type': 'refresh_token',
        'client_id': self.client_id,
        'client_secret': self.client_secret,
        'refresh_token': self.refresh_token,
        })
    return body

  def _generate_refresh_request_headers(self):
    """Generate the headers that will be used in the refresh request."""
    headers = {
        'content-type': 'application/x-www-form-urlencoded',
    }

    if self.user_agent is not None:
      headers['user-agent'] = self.user_agent

    return headers

  def _refresh(self, http_request):
    """Refreshes the access_token.

    This method first checks by reading the Storage object if available.
    If a refresh is still needed, it holds the Storage lock until the
    refresh is completed.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the refresh request.

    Raises:
      AccessTokenRefreshError: When the refresh fails.
    """
    if not self.store:
      self._do_refresh_request(http_request)
    else:
      self.store.acquire_lock()
      try:
        new_cred = self.store.locked_get()
        if (new_cred and not new_cred.invalid and
            new_cred.access_token != self.access_token):
          logger.info('Updated access_token read from Storage')
          self._updateFromCredential(new_cred)
        else:
          self._do_refresh_request(http_request)
      finally:
        self.store.release_lock()

  def _do_refresh_request(self, http_request):
    """Refresh the access_token using the refresh_token.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the refresh request.

    Raises:
      AccessTokenRefreshError: When the refresh fails.
    """
    body = self._generate_refresh_request_body()
    headers = self._generate_refresh_request_headers()

    logger.info('Refreshing access_token')
    resp, content = http_request(
        self.token_uri, method='POST', body=body, headers=headers)
    if resp.status == 200:
      # TODO(jcgregorio) Raise an error if loads fails?
      d = simplejson.loads(content)
      self.token_response = d
      self.access_token = d['access_token']
      self.refresh_token = d.get('refresh_token', self.refresh_token)
      if 'expires_in' in d:
        self.token_expiry = datetime.timedelta(
            seconds=int(d['expires_in'])) + datetime.datetime.utcnow()
      else:
        self.token_expiry = None
      if self.store:
        self.store.locked_put(self)
    else:
      # An {'error':...} response body means the token is expired or revoked,
      # so we flag the credentials as such.
      logger.info('Failed to retrieve access token: %s' % content)
      error_msg = 'Invalid response %s.' % resp['status']
      try:
        d = simplejson.loads(content)
        if 'error' in d:
          error_msg = d['error']
          self.invalid = True
          if self.store:
            self.store.locked_put(self)
      except Exception:
        pass
      raise AccessTokenRefreshError(error_msg)

  def _revoke(self, http_request):
    """Revokes the refresh_token and deletes the store if available.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the revoke request.
    """
    self._do_revoke(http_request, self.refresh_token)

  def _do_revoke(self, http_request, token):
    """Revokes the credentials and deletes the store if available.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the refresh request.
      token: A string used as the token to be revoked. Can be either an
        access_token or refresh_token.

    Raises:
      TokenRevokeError: If the revoke request does not return with a 200 OK.
    """
    logger.info('Revoking token')
    query_params = {'token': token}
    token_revoke_uri = _update_query_params(self.revoke_uri, query_params)
    resp, content = http_request(token_revoke_uri)
    if resp.status == 200:
      self.invalid = True
    else:
      error_msg = 'Invalid response %s.' % resp.status
      try:
        d = simplejson.loads(content)
        if 'error' in d:
          error_msg = d['error']
      except Exception:
        pass
      raise TokenRevokeError(error_msg)

    if self.store:
      self.store.delete()


class AccessTokenCredentials(OAuth2Credentials):
  """Credentials object for OAuth 2.0.

  Credentials can be applied to an httplib2.Http object using the
  authorize() method, which then signs each request from that object
  with the OAuth 2.0 access token. This set of credentials is for the
  use case where you have acquired an OAuth 2.0 access_token from
  another place such as a JavaScript client or another web
  application, and wish to use it from Python. Because only the
  access_token is present it can not be refreshed and will in time
  expire.

  AccessTokenCredentials objects may be safely pickled and unpickled.

  Usage:
    credentials = AccessTokenCredentials('<an access token>',
      'my-user-agent/1.0')
    http = httplib2.Http()
    http = credentials.authorize(http)

  Exceptions:
    AccessTokenCredentialsExpired: raised when the access_token expires or is
      revoked.
  """

  def __init__(self, access_token, user_agent, revoke_uri=None):
    """Create an instance of OAuth2Credentials

    This is one of the few types if Credentials that you should contrust,
    Credentials objects are usually instantiated by a Flow.

    Args:
      access_token: string, access token.
      user_agent: string, The HTTP User-Agent to provide for this application.
      revoke_uri: string, URI for revoke endpoint. Defaults to None; a token
        can't be revoked if this is None.
    """
    super(AccessTokenCredentials, self).__init__(
        access_token,
        None,
        None,
        None,
        None,
        None,
        user_agent,
        revoke_uri=revoke_uri)


  @classmethod
  def from_json(cls, s):
    data = simplejson.loads(s)
    retval = AccessTokenCredentials(
        data['access_token'],
        data['user_agent'])
    return retval

  def _refresh(self, http_request):
    raise AccessTokenCredentialsError(
        'The access_token is expired or invalid and can\'t be refreshed.')

  def _revoke(self, http_request):
    """Revokes the access_token and deletes the store if available.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the revoke request.
    """
    self._do_revoke(http_request, self.access_token)


class AssertionCredentials(OAuth2Credentials):
  """Abstract Credentials object used for OAuth 2.0 assertion grants.

  This credential does not require a flow to instantiate because it
  represents a two legged flow, and therefore has all of the required
  information to generate and refresh its own access tokens. It must
  be subclassed to generate the appropriate assertion string.

  AssertionCredentials objects may be safely pickled and unpickled.
  """

  @util.positional(2)
  def __init__(self, assertion_type, user_agent=None,
               token_uri=GOOGLE_TOKEN_URI,
               revoke_uri=GOOGLE_REVOKE_URI,
               **unused_kwargs):
    """Constructor for AssertionFlowCredentials.

    Args:
      assertion_type: string, assertion type that will be declared to the auth
        server
      user_agent: string, The HTTP User-Agent to provide for this application.
      token_uri: string, URI for token endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      revoke_uri: string, URI for revoke endpoint.
    """
    super(AssertionCredentials, self).__init__(
        None,
        None,
        None,
        None,
        None,
        token_uri,
        user_agent,
        revoke_uri=revoke_uri)
    self.assertion_type = assertion_type

  def _generate_refresh_request_body(self):
    assertion = self._generate_assertion()

    body = urllib.parse.urlencode({
        'assertion': assertion,
        'grant_type': 'urn:ietf:params:oauth:grant-type:jwt-bearer',
        })

    return body

  def _generate_assertion(self):
    """Generate the assertion string that will be used in the access token
    request.
    """
    _abstract()

  def _revoke(self, http_request):
    """Revokes the access_token and deletes the store if available.

    Args:
      http_request: callable, a callable that matches the method signature of
        httplib2.Http.request, used to make the revoke request.
    """
    self._do_revoke(http_request, self.access_token)


if HAS_CRYPTO:
  # PyOpenSSL and PyCrypto are not prerequisites for oauth2client, so if it is
  # missing then don't create the SignedJwtAssertionCredentials or the
  # verify_id_token() method.

  class SignedJwtAssertionCredentials(AssertionCredentials):
    """Credentials object used for OAuth 2.0 Signed JWT assertion grants.

    This credential does not require a flow to instantiate because it represents
    a two legged flow, and therefore has all of the required information to
    generate and refresh its own access tokens.

    SignedJwtAssertionCredentials requires either PyOpenSSL, or PyCrypto 2.6 or
    later. For App Engine you may also consider using AppAssertionCredentials.
    """

    MAX_TOKEN_LIFETIME_SECS = 3600 # 1 hour in seconds

    @util.positional(4)
    def __init__(self,
        service_account_name,
        private_key,
        scope,
        private_key_password='notasecret',
        user_agent=None,
        token_uri=GOOGLE_TOKEN_URI,
        revoke_uri=GOOGLE_REVOKE_URI,
        **kwargs):
      """Constructor for SignedJwtAssertionCredentials.

      Args:
        service_account_name: string, id for account, usually an email address.
        private_key: string, private key in PKCS12 or PEM format.
        scope: string or iterable of strings, scope(s) of the credentials being
          requested.
        private_key_password: string, password for private_key, unused if
          private_key is in PEM format.
        user_agent: string, HTTP User-Agent to provide for this application.
        token_uri: string, URI for token endpoint. For convenience
          defaults to Google's endpoints but any OAuth 2.0 provider can be used.
        revoke_uri: string, URI for revoke endpoint.
        kwargs: kwargs, Additional parameters to add to the JWT token, for
          example sub=joe@xample.org."""

      super(SignedJwtAssertionCredentials, self).__init__(
          None,
          user_agent=user_agent,
          token_uri=token_uri,
          revoke_uri=revoke_uri,
          )

      self.scope = util.scopes_to_string(scope)

      # Keep base64 encoded so it can be stored in JSON.
      self.private_key = base64.b64encode(private_key)

      self.private_key_password = private_key_password
      self.service_account_name = service_account_name
      self.kwargs = kwargs

    @classmethod
    def from_json(cls, s):
      data = simplejson.loads(s)
      retval = SignedJwtAssertionCredentials(
          data['service_account_name'],
          base64.b64decode(data['private_key']),
          data['scope'],
          private_key_password=data['private_key_password'],
          user_agent=data['user_agent'],
          token_uri=data['token_uri'],
          **data['kwargs']
          )
      retval.invalid = data['invalid']
      retval.access_token = data['access_token']
      return retval

    def _generate_assertion(self):
      """Generate the assertion that will be used in the request."""
      now = int(time.time())
      payload = {
          'aud': self.token_uri,
          'scope': self.scope,
          'iat': now,
          'exp': now + SignedJwtAssertionCredentials.MAX_TOKEN_LIFETIME_SECS,
          'iss': self.service_account_name
      }
      payload.update(self.kwargs)
      logger.debug(str(payload))

      private_key = base64.b64decode(self.private_key)
      return crypt.make_signed_jwt(crypt.Signer.from_string(
          private_key, self.private_key_password), payload)

  # Only used in verify_id_token(), which is always calling to the same URI
  # for the certs.
  _cached_http = httplib2.Http(MemoryCache())

  @util.positional(2)
  def verify_id_token(id_token, audience, http=None,
      cert_uri=ID_TOKEN_VERIFICATON_CERTS):
    """Verifies a signed JWT id_token.

    This function requires PyOpenSSL and because of that it does not work on
    App Engine.

    Args:
      id_token: string, A Signed JWT.
      audience: string, The audience 'aud' that the token should be for.
      http: httplib2.Http, instance to use to make the HTTP request. Callers
        should supply an instance that has caching enabled.
      cert_uri: string, URI of the certificates in JSON format to
        verify the JWT against.

    Returns:
      The deserialized JSON in the JWT.

    Raises:
      oauth2client.crypt.AppIdentityError if the JWT fails to verify.
    """
    if http is None:
      http = _cached_http

    resp, content = http.request(cert_uri)

    if resp.status == 200:
      certs = simplejson.loads(content)
      return crypt.verify_signed_jwt_with_certs(id_token, certs, audience)
    else:
      raise VerifyJwtTokenError('Status code: %d' % resp.status)


def _urlsafe_b64decode(b64string):
  # Guard against unicode strings, which base64 can't handle.
  b64string = b64string.encode('ascii')
  padded = b64string + '=' * (4 - len(b64string) % 4)
  return base64.urlsafe_b64decode(padded)


def _extract_id_token(id_token):
  """Extract the JSON payload from a JWT.

  Does the extraction w/o checking the signature.

  Args:
    id_token: string, OAuth 2.0 id_token.

  Returns:
    object, The deserialized JSON payload.
  """
  segments = id_token.split('.')

  if (len(segments) != 3):
    raise VerifyJwtTokenError(
      'Wrong number of segments in token: %s' % id_token)

  return simplejson.loads(_urlsafe_b64decode(segments[1]))


def _parse_exchange_token_response(content):
  """Parses response of an exchange token request.

  Most providers return JSON but some (e.g. Facebook) return a
  url-encoded string.

  Args:
    content: The body of a response

  Returns:
    Content as a dictionary object. Note that the dict could be empty,
    i.e. {}. That basically indicates a failure.
  """
  resp = {}
  try:
    resp = simplejson.loads(content)
  except Exception:
    # different JSON libs raise different exceptions,
    # so we just do a catch-all here
    resp = dict(parse_qsl(content))

  # some providers respond with 'expires', others with 'expires_in'
  if resp and 'expires' in resp:
    resp['expires_in'] = resp.pop('expires')

  return resp


@util.positional(4)
def credentials_from_code(client_id, client_secret, scope, code,
                          redirect_uri='postmessage', http=None,
                          user_agent=None, token_uri=GOOGLE_TOKEN_URI,
                          auth_uri=GOOGLE_AUTH_URI,
                          revoke_uri=GOOGLE_REVOKE_URI):
  """Exchanges an authorization code for an OAuth2Credentials object.

  Args:
    client_id: string, client identifier.
    client_secret: string, client secret.
    scope: string or iterable of strings, scope(s) to request.
    code: string, An authroization code, most likely passed down from
      the client
    redirect_uri: string, this is generally set to 'postmessage' to match the
      redirect_uri that the client specified
    http: httplib2.Http, optional http instance to use to do the fetch
    token_uri: string, URI for token endpoint. For convenience
      defaults to Google's endpoints but any OAuth 2.0 provider can be used.
    auth_uri: string, URI for authorization endpoint. For convenience
      defaults to Google's endpoints but any OAuth 2.0 provider can be used.
    revoke_uri: string, URI for revoke endpoint. For convenience
      defaults to Google's endpoints but any OAuth 2.0 provider can be used.

  Returns:
    An OAuth2Credentials object.

  Raises:
    FlowExchangeError if the authorization code cannot be exchanged for an
     access token
  """
  flow = OAuth2WebServerFlow(client_id, client_secret, scope,
                             redirect_uri=redirect_uri, user_agent=user_agent,
                             auth_uri=auth_uri, token_uri=token_uri,
                             revoke_uri=revoke_uri)

  credentials = flow.step2_exchange(code, http=http)
  return credentials


@util.positional(3)
def credentials_from_clientsecrets_and_code(filename, scope, code,
                                            message = None,
                                            redirect_uri='postmessage',
                                            http=None,
                                            cache=None):
  """Returns OAuth2Credentials from a clientsecrets file and an auth code.

  Will create the right kind of Flow based on the contents of the clientsecrets
  file or will raise InvalidClientSecretsError for unknown types of Flows.

  Args:
    filename: string, File name of clientsecrets.
    scope: string or iterable of strings, scope(s) to request.
    code: string, An authorization code, most likely passed down from
      the client
    message: string, A friendly string to display to the user if the
      clientsecrets file is missing or invalid. If message is provided then
      sys.exit will be called in the case of an error. If message in not
      provided then clientsecrets.InvalidClientSecretsError will be raised.
    redirect_uri: string, this is generally set to 'postmessage' to match the
      redirect_uri that the client specified
    http: httplib2.Http, optional http instance to use to do the fetch
    cache: An optional cache service client that implements get() and set()
      methods. See clientsecrets.loadfile() for details.

  Returns:
    An OAuth2Credentials object.

  Raises:
    FlowExchangeError if the authorization code cannot be exchanged for an
     access token
    UnknownClientSecretsFlowError if the file describes an unknown kind of Flow.
    clientsecrets.InvalidClientSecretsError if the clientsecrets file is
      invalid.
  """
  flow = flow_from_clientsecrets(filename, scope, message=message, cache=cache,
                                 redirect_uri=redirect_uri)
  credentials = flow.step2_exchange(code, http=http)
  return credentials


class OAuth2WebServerFlow(Flow):
  """Does the Web Server Flow for OAuth 2.0.

  OAuth2WebServerFlow objects may be safely pickled and unpickled.
  """

  @util.positional(4)
  def __init__(self, client_id, client_secret, scope,
               redirect_uri=None,
               user_agent=None,
               auth_uri=GOOGLE_AUTH_URI,
               token_uri=GOOGLE_TOKEN_URI,
               revoke_uri=GOOGLE_REVOKE_URI,
               **kwargs):
    """Constructor for OAuth2WebServerFlow.

    The kwargs argument is used to set extra query parameters on the
    auth_uri. For example, the access_type and approval_prompt
    query parameters can be set via kwargs.

    Args:
      client_id: string, client identifier.
      client_secret: string client secret.
      scope: string or iterable of strings, scope(s) of the credentials being
        requested.
      redirect_uri: string, Either the string 'urn:ietf:wg:oauth:2.0:oob' for
        a non-web-based application, or a URI that handles the callback from
        the authorization server.
      user_agent: string, HTTP User-Agent to provide for this application.
      auth_uri: string, URI for authorization endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      token_uri: string, URI for token endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      revoke_uri: string, URI for revoke endpoint. For convenience
        defaults to Google's endpoints but any OAuth 2.0 provider can be used.
      **kwargs: dict, The keyword arguments are all optional and required
                        parameters for the OAuth calls.
    """
    self.client_id = client_id
    self.client_secret = client_secret
    self.scope = util.scopes_to_string(scope)
    self.redirect_uri = redirect_uri
    self.user_agent = user_agent
    self.auth_uri = auth_uri
    self.token_uri = token_uri
    self.revoke_uri = revoke_uri
    self.params = {
        'access_type': 'offline',
        'response_type': 'code',
    }
    self.params.update(kwargs)

  @util.positional(1)
  def step1_get_authorize_url(self, redirect_uri=None):
    """Returns a URI to redirect to the provider.

    Args:
      redirect_uri: string, Either the string 'urn:ietf:wg:oauth:2.0:oob' for
        a non-web-based application, or a URI that handles the callback from
        the authorization server. This parameter is deprecated, please move to
        passing the redirect_uri in via the constructor.

    Returns:
      A URI as a string to redirect the user to begin the authorization flow.
    """
    if redirect_uri is not None:
      logger.warning(('The redirect_uri parameter for'
          'OAuth2WebServerFlow.step1_get_authorize_url is deprecated. Please'
          'move to passing the redirect_uri in via the constructor.'))
      self.redirect_uri = redirect_uri

    if self.redirect_uri is None:
      raise ValueError('The value of redirect_uri must not be None.')

    query_params = {
        'client_id': self.client_id,
        'redirect_uri': self.redirect_uri,
        'scope': self.scope,
    }
    query_params.update(self.params)
    return _update_query_params(self.auth_uri, query_params)

  @util.positional(2)
  def step2_exchange(self, code, http=None):
    """Exhanges a code for OAuth2Credentials.

    Args:
      code: string or dict, either the code as a string, or a dictionary
        of the query parameters to the redirect_uri, which contains
        the code.
      http: httplib2.Http, optional http instance to use to do the fetch

    Returns:
      An OAuth2Credentials object that can be used to authorize requests.

    Raises:
      FlowExchangeError if a problem occured exchanging the code for a
      refresh_token.
    """

    if not (isinstance(code, str) or isinstance(code, str)):
      if 'code' not in code:
        if 'error' in code:
          error_msg = code['error']
        else:
          error_msg = 'No code was supplied in the query parameters.'
        raise FlowExchangeError(error_msg)
      else:
        code = code['code']

    body = urllib.parse.urlencode({
        'grant_type': 'authorization_code',
        'client_id': self.client_id,
        'client_secret': self.client_secret,
        'code': code,
        'redirect_uri': self.redirect_uri,
        'scope': self.scope,
        })
    headers = {
        'content-type': 'application/x-www-form-urlencoded',
    }

    if self.user_agent is not None:
      headers['user-agent'] = self.user_agent

    if http is None:
      http = httplib2.Http()

    resp, content = http.request(self.token_uri, method='POST', body=body,
                                 headers=headers)
    d = _parse_exchange_token_response(content)
    if resp.status == 200 and 'access_token' in d:
      access_token = d['access_token']
      refresh_token = d.get('refresh_token', None)
      token_expiry = None
      if 'expires_in' in d:
        token_expiry = datetime.datetime.utcnow() + datetime.timedelta(
            seconds=int(d['expires_in']))

      if 'id_token' in d:
        d['id_token'] = _extract_id_token(d['id_token'])

      logger.info('Successfully retrieved access token')
      return OAuth2Credentials(access_token, self.client_id,
                               self.client_secret, refresh_token, token_expiry,
                               self.token_uri, self.user_agent,
                               revoke_uri=self.revoke_uri,
                               id_token=d.get('id_token', None),
                               token_response=d)
    else:
      logger.info('Failed to retrieve access token: %s' % content)
      if 'error' in d:
        # you never know what those providers got to say
        error_msg = str(d['error'])
      else:
        error_msg = 'Invalid response: %s.' % str(resp.status)
      raise FlowExchangeError(error_msg)


@util.positional(2)
def flow_from_clientsecrets(filename, scope, redirect_uri=None,
                            message=None, cache=None):
  """Create a Flow from a clientsecrets file.

  Will create the right kind of Flow based on the contents of the clientsecrets
  file or will raise InvalidClientSecretsError for unknown types of Flows.

  Args:
    filename: string, File name of client secrets.
    scope: string or iterable of strings, scope(s) to request.
    redirect_uri: string, Either the string 'urn:ietf:wg:oauth:2.0:oob' for
      a non-web-based application, or a URI that handles the callback from
      the authorization server.
    message: string, A friendly string to display to the user if the
      clientsecrets file is missing or invalid. If message is provided then
      sys.exit will be called in the case of an error. If message in not
      provided then clientsecrets.InvalidClientSecretsError will be raised.
    cache: An optional cache service client that implements get() and set()
      methods. See clientsecrets.loadfile() for details.

  Returns:
    A Flow object.

  Raises:
    UnknownClientSecretsFlowError if the file describes an unknown kind of Flow.
    clientsecrets.InvalidClientSecretsError if the clientsecrets file is
      invalid.
  """
  try:
    client_type, client_info = clientsecrets.loadfile(filename, cache=cache)
    if client_type in (clientsecrets.TYPE_WEB, clientsecrets.TYPE_INSTALLED):
      constructor_kwargs = {
          'redirect_uri': redirect_uri,
          'auth_uri': client_info['auth_uri'],
          'token_uri': client_info['token_uri'],
      }
      revoke_uri = client_info.get('revoke_uri')
      if revoke_uri is not None:
        constructor_kwargs['revoke_uri'] = revoke_uri
      return OAuth2WebServerFlow(
          client_info['client_id'], client_info['client_secret'],
          scope, **constructor_kwargs)

  except clientsecrets.InvalidClientSecretsError:
    if message:
      sys.exit(message)
    else:
      raise
  else:
    raise UnknownClientSecretsFlowError(
        'This OAuth 2.0 flow is unsupported: %r' % client_type)



# Generated by Django 4.1.4 on 2023-11-15 19:46

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('account', '0014_alter_notificationmemberrecord_member'),
    ]

    operations = [
        migrations.AddField(
            model_name='memberdevice',
            name='buid',
            field=models.CharField(blank=True, db_index=True, default=None, max_length=128, null=True),
        ),
    ]



# Generated by Django 4.1.4 on 2023-01-27 04:57

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('metrics', '0002_metrics_k11_metrics_k12_metrics_k13_metrics_k14_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='metrics',
            name='expires',
            field=models.DateTimeField(db_index=True, default=None, null=True),
        ),
    ]



from objict import objict
from rest import settings
import requests

IPINFO_TOKEN = settings.get("GEOIP_IPINFO_TOKEN", "9ff8b6d9dd39b0")


def getLocation(ip):
    """
    {
    "ip": "12.173.248.130",
    "city": "Gatlinburg",
    "region": "Tennessee",
    "country": "US",
    "loc": "35.7145,-83.5119",
    "org": "AS7018 AT&T Services, Inc.",
    "postal": "37738",
    "timezone": "America/New_York",
    "readme": "https://ipinfo.io/missingauth"
    }
    """
    params = {"token": IPINFO_TOKEN}
    url = "https://ipinfo.io/{}/json".format(ip)
    res = requests.get(url, params=params)
    if res.status_code != 200:
        print(res.text)
        return None
    data = objict.fromdict(res.json())
    data.state = data.region
    data.isp = data.org

    loc = objict(provider="ipinfo")
    loc.isp = data.isp
    loc.ip = data.ip
    loc.hostname = data.hostname
    loc.ip_type = ""
    data.lat, data.lon = data.loc.split(',')
    loc.latitude = data.lat
    loc.longitude = data.lon
    loc.continent = ""
    loc.region = data.region
    loc.country = data.country
    loc.countryCode = data.country
    loc.state = data.region
    loc.city = data.city
    loc.business = data.org
    loc.website = ""
    loc.zipcode = data.postal
    return loc



from ..render_utils import *
from ..exceptions import *
from .. import render_utils

import os
import tempfile
import copy
from PIL import Image


from rest.log import getLogger
logger = getLogger("medialib", filename="medialib.log")


PresetConfig = [{
    'name':		'Image Transcode',
    'description':	'Transcode image to alternate size/quality/container',
    'stage':	50,
    'short_name':	'image_transcode',
    'default_use':	'image',
    'applicable_kinds': 'I',
    'parameters': {
        'image_container': {
            'description':	'Output image container extension',
            'required':	True,
            'kind':		'C',
            'configurable':	True,
            'choices':	'gif,jpg,png',
            'value':	'jpg',
        },
        'detect_black': {
            'description':	'detect black bounding areas',
            'required':	False,
            'kind':		'B',
            'configurable':	True,
            'value':	1,
        },
        'quality': {
            'description':	'Image quality, 1-100',
            'required':	False,
            'kind':		'I',
            'configurable':	True,
            'value':	75,
        },
    },
}]
PresetConfig[0]['parameters'].update(copy.deepcopy(render_utils.PresetParameters))

PresetConfig.append(copy.deepcopy(PresetConfig[0]))
PresetConfig[1].update({
    'name':		'Image Thumbnail',
    'description':	'transcode image to thumbnail',
    'stage':	20,
    'short_name':	'thumbnail',
})
render_utils.update_params(PresetConfig[1],
    maxwidth = 128,
    maxaspect = 1,
    quality = 95,
    minheight = 32,
    minwidth = 32,
)

PresetConfig.append(copy.deepcopy(PresetConfig[0]))
PresetConfig[2].update({
    'name':		'Image Thumbnail (zoomed)',
    'description':	'transcode image to zoomed-in thumbnail of specific size',
    'stage':	20,
    'short_name':	'thumbnail_zoomed',
})
render_utils.update_params(PresetConfig[2],
    quality = 95,
)


container_settings = {
    'jpg': {
        'mode': 'RGBA',
        'kind': 'JPEG',
    },
    'gif': {
        'mode': 'RGBA',
        'kind': 'GIF',
    },
    'png': {
        'mode': 'RGBA',
        'kind': 'GIF',
    },
}
        

def render(item, renditiondef, params):
    logger.info("image_transacoding...")
    if params.get('image_container', '') in container_settings:
        container = container_settings[params['image_container']]
    else:
        raise RenderError('%s: Unknown image container: %s' % (renditiondef.name, params.get('image_container', 'Undefined')))

    original = get_rendition(item, 'still')
    params['no_threads'] = True
    fp = None
    logger.info("image_transacoding 2...")
    try:
        im = Image.open(original.filename)
        orig_kind = im.format
        if orig_kind:
            orig_kind = orig_kind.lower()

        im.draft(container['mode'], im.size)
        
        im = adjust_exif(im)
        
        size = im.size
        if 'detect_black' in params and params['detect_black']:
            bound = Image.eval(im, lambda p: p>5 and p or 0).getbbox()
            
            if bound:
                bound_size = (bound[2]-bound[0], bound[3]-bound[1])
                
                if bound_size == size:
                    # original size and bound are the same - ignore boundary
                    bound = None
                else:
                    size = bound_size
        else:
            bound = None
        # print "presize: ",
        # print params
        logger.info("image params..", params)
        params = calc_size(params, size[0], size[1], even=False, rotate=original.get_meta('rotate'))
        logger.info("image post params..", params)
        # print "rendering image: ",
        # print params

        if not bound:
            pass
        elif 'crop' in params:
            params['crop'] = (params['crop'][0]+bound[0], params['crop'][1]+bound[1], params['crop'][2]+bound[0], params['crop'][3]+bound[1], )
        else:
            params['crop'] = bound

        if 'crop' in params:
            im = im.crop(params['crop'])
        if 'width' in params and 'height' in params:
            if size[0] > params['width']*3 and size[1] > params['height']*3:
                # do 2-pass resize - fast low quality resize
                im.thumbnail((params['width']*3, params['height']*3))
            logger.info("image resizing..")
            # do slow high quality resize
            im = im.resize((params['width'], params['height']), Image.ANTIALIAS)
        
        if im.mode != container['mode']:
            im = im.convert(container['mode'])

        saveopts = {}
        if 'quality' in params:
            saveopts['quality'] = params['quality']

        # ians keep the original format if GIF or PNG to keep transparency
        kind = container["kind"].lower()

        if orig_kind != kind and orig_kind in ["gif", "png"]:
            kind = orig_kind 
        # print container
        # print "original format: {0}... new format: {1}".format(orig_kind, kind)
        # fp = tempfile.NamedTemporaryFile(suffix=('.' + params.get('image_container', 'jpg')), prefix="t_transcode_")
        fp = tempfile.NamedTemporaryFile(suffix=('.' + kind), prefix="t_transcode_")
        dstfile = fp.name
        if im.mode == "RGBA" and kind not in ["gif", "png"]:
            # im = im.convert("RGB")
            kind = "png"

        im.save(dstfile, kind, **saveopts)
        logger.info("image saved: {}".format(dstfile))
        new_rendition(item, renditiondef, fp, kind='I', width=im.size[0], height=im.size[1])
        if hasattr(im, "close"):
            im.close()
        fp.close()

    except (IOError, SyntaxError, ValueError, EOFError, AttributeError, IndexError, CmdError) as err:
        logger.error("image_transcode")
        try:
            if fp:
                fp.close()
        except:
            pass
        raise RenderError('%s: %s' % (renditiondef.name, str(err)), stack=err, show_stack=True)

def adjust_exif(im):
    # fix any camera rotations
    try:
        info = im._getexif()
    except (AttributeError, KeyError):
        return im
    if info and 0x0112 in info and info[0x0112]:
        orient = info[0x0112]
        if orient == 1: # A-OK
            pass
        if orient == 2: # mirror horz
            im = im.transpose(FLIP_LEFT_RIGHT)
        elif orient == 3: # rotated 180
            im = im.transpose(Image.ROTATE_180)
        elif orient == 4: # mirror vert
            im = im.transpose(Image.FLIP_TOP_BOTTOM)
        elif orient == 5: # mirror horz rotated 90 ccw
            im = im.transpose(Image.ROTATE_90).transpose(FLIP_LEFT_RIGHT)
        elif orient == 6: # rotated 90 cw
            im = im.transpose(Image.ROTATE_270)
        elif orient == 7: # mirror horz rotated 90 cw
            im = im.transpose(Image.ROTATE_270).transpose(FLIP_LEFT_RIGHT)
        elif orient == 8: # rotated 90 ccw
            im = im.transpose(Image.ROTATE_90)
    
    return im



###############################################################################
##
##  Copyright 2011-2013 Tavendo GmbH
##
##  Note:
##
##  This code is a Python implementation of the algorithm
##
##            "Flexible and Economical UTF-8 Decoder"
##
##  by Bjoern Hoehrmann
##
##       bjoern@hoehrmann.de
##       http://bjoern.hoehrmann.de/utf-8/decoder/dfa/
##
##  Licensed under the Apache License, Version 2.0 (the "License");
##  you may not use this file except in compliance with the License.
##  You may obtain a copy of the License at
##
##      http://www.apache.org/licenses/LICENSE-2.0
##
##  Unless required by applicable law or agreed to in writing, software
##  distributed under the License is distributed on an "AS IS" BASIS,
##  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
##  See the License for the specific language governing permissions and
##  limitations under the License.
##
###############################################################################


## use Cython implementation of UTF8 validator if available
##
try:
    from wsaccel.utf8validator import Utf8Validator
except:
    ## fallback to pure Python implementation

    class Utf8Validator:
        """
        Incremental UTF-8 validator with constant memory consumption (minimal
        state).

        Implements the algorithm "Flexible and Economical UTF-8 Decoder" by
        Bjoern Hoehrmann (http://bjoern.hoehrmann.de/utf-8/decoder/dfa/).
        """

        ## DFA transitions
        UTF8VALIDATOR_DFA = [
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  # 00..1f
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  # 20..3f
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  # 40..5f
            0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,  # 60..7f
            1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,9,  # 80..9f
            7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,7,  # a0..bf
            8,8,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,2,  # c0..df
            0xa,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x3,0x4,0x3,0x3,  # e0..ef
            0xb,0x6,0x6,0x6,0x5,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,0x8,  # f0..ff
            0x0,0x1,0x2,0x3,0x5,0x8,0x7,0x1,0x1,0x1,0x4,0x6,0x1,0x1,0x1,0x1,  # s0..s0
            1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,1,0,1,1,1,1,1,0,1,0,1,1,1,1,1,1,  # s1..s2
            1,2,1,1,1,1,1,2,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,  # s3..s4
            1,2,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,1,1,1,1,1,1,3,1,3,1,1,1,1,1,1,  # s5..s6
            1,3,1,1,1,1,1,3,1,3,1,1,1,1,1,1,1,3,1,1,1,1,1,1,1,1,1,1,1,1,1,1,  # s7..s8
        ]

        UTF8_ACCEPT = 0
        UTF8_REJECT = 1

        def __init__(self):
            self.reset()

        def decode(self, b):
            """
            Eat one UTF-8 octet, and validate on the fly.

            Returns UTF8_ACCEPT when enough octets have been consumed, in which case
            self.codepoint contains the decoded Unicode code point.

            Returns UTF8_REJECT when invalid UTF-8 was encountered.

            Returns some other positive integer when more octets need to be eaten.
            """
            type = Utf8Validator.UTF8VALIDATOR_DFA[b]

            if self.state != Utf8Validator.UTF8_ACCEPT:
                self.codepoint = (b & 0x3f) | (self.codepoint << 6)
            else:
                self.codepoint = (0xff >> type) & b

            self.state = Utf8Validator.UTF8VALIDATOR_DFA[256 + self.state * 16 + type]

            return self.state

        def reset(self):
            """
            Reset validator to start new incremental UTF-8 decode/validation.
            """
            self.state = Utf8Validator.UTF8_ACCEPT
            self.codepoint = 0
            self.i = 0

        def validate(self, ba):
            """
            Incrementally validate a chunk of bytes provided as string.

            Will return a quad (valid?, endsOnCodePoint?, currentIndex, totalIndex).

            As soon as an octet is encountered which renders the octet sequence
            invalid, a quad with valid? == False is returned. currentIndex returns
            the index within the currently consumed chunk, and totalIndex the
            index within the total consumed sequence that was the point of bail out.
            When valid? == True, currentIndex will be len(ba) and totalIndex the
            total amount of consumed bytes.
            """

            l = len(ba)

            for i in range(l):
                ## optimized version of decode(), since we are not interested in actual code points

                self.state = Utf8Validator.UTF8VALIDATOR_DFA[256 + (self.state << 4) + Utf8Validator.UTF8VALIDATOR_DFA[ord(ba[i])]]

                if self.state == Utf8Validator.UTF8_REJECT:
                    self.i += i
                    return False, False, i, self.i

            self.i += l

            return True, self.state == Utf8Validator.UTF8_ACCEPT, l, self.i



from ..exceptions import *
from ..render_utils import RenderProc, PipeThread

from django import db

import select
import fcntl
import os
import sys

class RtmpSinkRun(PipeThread):
    def read(self, infd, outf):
        select.select((infd,), (), ())
        return os.read(infd, 10240)
    
    def write(self, data, infd, outf):
        return outf.write(data)
        
def run(url, fname):
    from .rtmp import rtmp
    (rp, wp) = os.pipe()
    pid = os.fork()

    if pid == 0:
        global tdata
        os.dup2(wp, sys.stdout.fileno())
        os.dup2(wp, sys.stderr.fileno())
        os.close(rp)
        os.close(wp)
        
        fd = os.open(fname, os.O_RDONLY)
        fcntl.fcntl(fd, fcntl.F_SETFL, fcntl.fcntl(fd, fcntl.F_GETFL) | os.O_NONBLOCK)

        db.connection.close()

        out = rtmp.sink(str(url))

        RtmpSinkRun(fd, out, delay_error=True)

        out.close()
        os.close(fd)
        os._exit(0)
    else:
        os.close(wp)
        return RenderProc(pid, errfile=rp)




# Copyright (C) 2013 Google Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for making samples.

Consolidates a lot of code commonly repeated in sample applications.
"""

__author__ = 'jcgregorio@google.com (Joe Gregorio)'
__all__ = ['init']


import argparse
import httplib2
import os

from medialib.youtube.apiclient import discovery
from medialib.youtube.oauth2client import client
from medialib.youtube.oauth2client import file
from medialib.youtube.oauth2client import tools


def init(argv, name, version, doc, filename, scope=None, parents=[]):
  """A common initialization routine for samples.

  Many of the sample applications do the same initialization, which has now
  been consolidated into this function. This function uses common idioms found
  in almost all the samples, i.e. for an API with name 'apiname', the
  credentials are stored in a file named apiname.dat, and the
  client_secrets.json file is stored in the same directory as the application
  main file.

  Args:
    argv: list of string, the command-line parameters of the application.
    name: string, name of the API.
    version: string, version of the API.
    doc: string, description of the application. Usually set to __doc__.
    file: string, filename of the application. Usually set to __file__.
    parents: list of argparse.ArgumentParser, additional command-line flags.
    scope: string, The OAuth scope used.

  Returns:
    A tuple of (service, flags), where service is the service object and flags
    is the parsed command-line flags.
  """
  if scope is None:
    scope = 'https://www.googleapis.com/auth/' + name

  # Parser command-line arguments.
  parent_parsers = [tools.argparser]
  parent_parsers.extend(parents)
  parser = argparse.ArgumentParser(
      description=doc,
      formatter_class=argparse.RawDescriptionHelpFormatter,
      parents=parent_parsers)
  flags = parser.parse_args(argv[1:])

  # Name of a file containing the OAuth 2.0 information for this
  # application, including client_id and client_secret, which are found
  # on the API Access tab on the Google APIs
  # Console <http://code.google.com/apis/console>.
  client_secrets = os.path.join(os.path.dirname(filename),
                                'client_secrets.json')

  # Set up a Flow object to be used if we need to authenticate.
  flow = client.flow_from_clientsecrets(client_secrets,
      scope=scope,
      message=tools.message_if_missing(client_secrets))

  # Prepare credentials, and authorize HTTP object with them.
  # If the credentials don't exist or are invalid run through the native client
  # flow. The Storage object will ensure that if successful the good
  # credentials will get written back to a file.
  storage = file.Storage(name + '.dat')
  credentials = storage.get()
  if credentials is None or credentials.invalid:
    credentials = tools.run_flow(flow, storage, flags)
  http = credentials.authorize(http = httplib2.Http())

  # Construct a service object via the discovery service.
  service = discovery.build(name, version, http=http)
  return (service, flags)



from ..engines import ffmpeg, anigif
from ..render_utils import *
from ..exceptions import *
from .. import render_utils

import os
import shutil
import tempfile
import glob
import time
import copy
from PIL import Image

PresetConfig = {
    'name':		'Animated Thumbnail',
    'description':	'Animated GIF Thumbnail',
    'stage':	101,
    'use':		'thumbnail-animated',
    'applicable_kinds': 'VLAS',
    'parameters': {
        'frame': {
            'description':	'Number of frames to encode',
            'required':	False,
            'kind':		'I',
            'configurable':	True,
        },
        'frame_rate': {
            'description':	'Frame rate',
            'required':	True,
            'kind':		'I',
            'configurable':	True,
            'value':	4,
        },
        'seek_time': {
            'description':	'Seek from start (sec)',
            'required':	True,
            'kind':		'F',
            'configurable':	True,
        },
        'duration': {
            'description':	'Duration',
            'required':	False,
            'kind':		'F',
            'configurable':	True,
        },
    },
}
PresetConfig['parameters'].update(copy.deepcopy(render_utils.PresetParameters))

def render(item, renditiondef, params):
    fp = tempfile.NamedTemporaryFile(suffix=('.' + params.get('image_container', 'gif')), prefix="t_anigif_")
    dstfile = fp.name
    tmpdir = tempfile.mkdtemp()
    prevdir = os.getcwd()
    os.chdir(tmpdir)
    
    params['pass'] = 1

    original = get_rendition(item, 'original')
    params['no_threads'] = True
    if not 'frame' in params:
        params['frame'] = 20
    
    try:
        ffmpeg_params = params.copy()
        try:
            delattr(ffmpeg_params, 'width')
        except AttributeError:
            pass
        try:
            delattr(ffmpeg_params, 'height')
        except AttributeError:
            pass
        owidth = original.width
        oheight = original.height
        rotate=original.get_meta('rotate')
        if rotate:
            ffmpeg_params['rotate'] = rotate
            if rotate == 90 or rotate == 270:
                x = oheight
                oheight = owidth
                owidth = x
        ffmpeg.run(tmpdir + "/animate_%06d.png", original.filename, ffmpeg_params)

        params = calc_size(params, owidth, oheight)

        afiles = glob.glob(tmpdir + "/animate_*.png")
        afiles.sort()
        
        images = []
        for inf in afiles:
            im = Image.open(inf)

            if 'crop' in params:
                im = im.crop(params['crop'])
            if 'width' in params and 'height' in params:
                im.thumbnail((params['width'], params['height']))
        
            images.append(im)

        durations = [0.01] * len(afiles)
        durations[-1] = 1
        anigif.run(fp, images, durations, 0)

        new_rendition(item, renditiondef, fp, kind='N', width=params.get('width', None), height=params.get('height', None))
    except (IOError, SyntaxError, ValueError, EOFError, AttributeError, IndexError, CmdError) as err:
        os.chdir(prevdir)
        raise RenderError('%s: %s' % (renditiondef.name, str(err)), stack=err, show_stack=True)
    finally:
        os.chdir(prevdir)
        shutil.rmtree(tmpdir)







from django.db import models
from django.conf import settings


from account.models import User, Member, Group, MemberFeed
from medialib.models import MediaItem, MediaLibrary
from location.models import GeoIP, Address

from rest import helpers as rest_helpers
from rest.middleware import get_request
from rest import RemoteEvents
from rest.models import RestModel, MetaDataModel, MetaDataBase

class Product(models.Model, RestModel, MetaDataModel):
    class RestMeta:
        GRAPHS = {
            "base": {
                "graphs": {
                    "current":"default",
                    "beta":"default",
                },
                "exclude":["library"]
            },
            "default": {
                "graphs": {
                    "current":"default",
                    "beta":"default",
                },
                "recurse_into":[("getReleases", "versions")],
                "exclude":["library"]
            }
        }
    oid = models.CharField(max_length=250, db_index=True, blank=True, default="")
    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True)

    owner = models.ForeignKey(Member, related_name="+", on_delete=models.CASCADE)
    group = models.ForeignKey(Group, related_name="+", blank=True, null=True, default=None, on_delete=models.CASCADE)

    archived = models.BooleanField(default=False, blank=True)

    name = models.CharField(max_length=200)
    description = models.TextField(blank=True, null=True, default=None)
    kind = models.CharField(max_length=64, blank=True, null=True, default=None)
    is_public = models.BooleanField(default=False, blank=True)

    current = models.ForeignKey("Release", related_name="+", blank=True, null=True, default=None, on_delete=models.CASCADE)
    beta = models.ForeignKey("Release", related_name="+", blank=True, null=True, default=None, on_delete=models.CASCADE)

    library = models.ForeignKey(MediaLibrary, blank=True, null=True, default=None, on_delete=models.CASCADE)

    def getReleases(self):
        return self.releases.all().order_by("-created")[:10]

    def getLibrary(self, commit=True):
        if not self.library:  # object is being created, thus no primary key field yet
            library = MediaLibrary(owner=self.owner, name="Releases")
            library.save()
            self.library = library
            if commit:
                self.save()
        return self.library

    def save(self, *args, **kwargs):
        self.getLibrary(False)
        super(Product, self).save(*args, **kwargs)

    def __unicode__(self):
        return "{0} - {1}".format(self.name, self.kind)


class ProductMetaData(MetaDataBase):
    parent = models.ForeignKey(Product, related_name="properties", on_delete=models.CASCADE)

class Release(models.Model, RestModel, MetaDataModel):
    class RestMeta:
        GRAPHS = {
            "default": {
                "exclude":["media"],
                "extra":["download_url"]
            }
        }

        POST_SAVE_FIELDS = ["media"]


    created = models.DateTimeField(auto_now_add=True, editable=False)
    modified = models.DateTimeField(auto_now=True)
    owner = models.ForeignKey(Member, related_name="+", on_delete=models.CASCADE)

    product = models.ForeignKey(Product, related_name="releases", on_delete=models.CASCADE)

    notes = models.TextField(blank=True, null=True, default=None)
    version_num = models.IntegerField(db_index=True, default=1)
    version_str = models.CharField(max_length=32, blank=True, null=True)

    media = models.ForeignKey(MediaItem, on_delete=models.PROTECT, blank=True, null=True, default=None)

    def makeCurrent(self):
        self.product.current = self
        self.product.save()

    def makeBeta(self):
        self.product.beta = self
        self.product.save()

    def download_url(self):
        if self.media:
            return self.media.url()
        return None

    def set_media(self, value, commit=False):
        if value is None:
            return

        if value.startswith("http"):
            media = MediaItem.CreateFromURL(value, self.owner, library=self.product.getLibrary(), kind=None)
            self.media = media
            if commit:
                self.save()
            return media

        request = get_request()
        name = request.DATA.get("media_name", "test.txt")
        kind = MediaItem.guessMediaKind(value)
        media = MediaItem(library=self.product.getLibrary(), name=name, owner=self.product.owner, kind=kind, base64_data=value)
        media.save()
        self.media = media

    # upload__ called for files
    def upload__media(self, value, name):
        if value is None:
            return

        kind = MediaItem.guessMediaKind(value)
        media = MediaItem(library=self.product.getLibrary(), name=name, owner=self.product.owner, kind=kind, newfile=value)
        media.save()
        self.media = media

    def __unicode__(self):
        return "{0} - {1}".format(self.product.name, self.version_str)






from django.db import models
from django.conf import settings

from rest import models as rm
from rest import helpers as rh


class Ticket(models.Model, rm.RestModel, rm.MetaDataModel):
    class RestMeta:
        SEARCH_FIELDS = ["title", "description"]
        CAN_DELETE = True
        VIEW_PERMS = ["view_tickets"]
        GRAPHS = {
            "default": {
                "extra": ["metadata"],
                "graphs": {
                    "group": "basic",
                    "created_by": "basic",
                    "assigned_to": "basic",
                    "incident": "basic"
                },
            },
            "detailed": {
                "graphs": {
                    "group": "basic",
                    "created_by": "basic",
                    "assigned_to": "basic",
                    "incident": "basic",
                    "generic__component": "basic"
                },
            },
        }

    created = models.DateTimeField(auto_now_add=True)
    modified = models.DateTimeField(auto_now=True)

    group = models.ForeignKey(
        "account.Group", on_delete=models.SET_NULL, 
        related_name="tickets",
        null=True, default=None)
    assigned_to = models.ForeignKey(
        "account.Member", on_delete=models.SET_NULL, 
        related_name="tickets",
        null=True, default=None)
    created_by = models.ForeignKey(
        "account.Member", on_delete=models.CASCADE,
        related_name="+",
        null=True, default=None)

    # if category is null then this will run on all events?
    category = models.CharField(max_length=200, db_index=True)
    priority = models.IntegerField(default=10, db_index=True)  # 1-10, 1 being the highest

    title = models.CharField(max_length=200)
    description = models.TextField()

    status = models.CharField(max_length=32, default=None, null=True, db_index=True)
    state = models.IntegerField(default=0)  # how many incidents before firing action

    incident = models.ForeignKey(
        "incident.Incident", null=True, default=None, 
        related_name="tickets", on_delete=models.SET_NULL)

    component = models.CharField(max_length=200, null=True, default=None, db_index=True)
    component_id = models.IntegerField(null=True, blank=True, default=None, db_index=True)



import time
from rest.jwtoken import JWToken
from rest import UberDict
from rest.crypto import util
from django.core.handlers.wsgi import WSGIRequest
from django.apps import apps

from ws4redis import settings as private_settings
from ws4redis.redis import RedisStore, RedisMessage

from rest.log import getLogger
logger = getLogger("async", filename="async.log")

MODEL_CACHE = dict()  # caching of app.Model for faster access

ALLOW_ANY_FACILITY = not private_settings.WS4REDIS_FACILITIES


def getRemoteIP(request):
    x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
    if x_forwarded_for:
        ip = x_forwarded_for.split(',')[0]
    else:
        ip = request.META.get('REMOTE_ADDR')
    return ip


class WebsocketConnection():
    def __init__(self, server, environ, start_response):
        self.server = server
        self.request = WSGIRequest(environ)
        self.ip = getRemoteIP(self.request)
        self.ua = self.request.META.get('HTTP_USER_AGENT', '')
        self.facility = self.request.path_info.replace(private_settings.WEBSOCKET_URL, '', 1)
        self.credentials = UberDict()
        self.listening_fds = None
        self.redis = RedisStore(server._redis_connection)
        self.websocket = server.upgrade_websocket(environ, start_response)
        self.last_beat = time.time()
        self.last_msg = None
        self._heart_beat = util.toString(private_settings.WS4REDIS_HEARTBEAT)
        self.debug = private_settings.WS4REDIS_LOG_DEBUG

    def refreshFDs(self):
        if len(self.listening_fds) == 1:
            sub_sd = self.redis.get_file_descriptor()
            if sub_sd:
                self.listening_fds.append(sub_sd)

    def on_auth(self, msg):
        if msg.kind == "jwt":
            self.on_auth_jwt(msg)
            return
        # check our other auth mechanisms
        auther = self.getAuthenticator(msg.kind)
        if auther is None or not hasattr(auther, "authWS4RedisConnection"):
            logger.error("invalid auth", msg)
            self.sendToWS(msg.channel, dict(error="invalid auth kind", code=500))
            return

        self.credentials = auther.authWS4RedisConnection(msg)
        if self.credentials is None or self.credentials.pk is None:
            logger.error("invalid credentials for", msg, self.credentials)
            self.sendToWS(msg.channel, dict(error="invalid credentials", code=401))
            return
        self.on_authenticated()

    def on_auth_jwt(self, msg):
        token = JWToken(msg.token)
        if token.payload is None:
            raise Exception("invalid auth token")
        Member = apps.get_model("account", "Member")
        member = Member.objects.filter(pk=token.payload.user_id).last()
        if member is None:
            raise Exception("invalid user token")
        token.key = member.security_token
        if not token.is_valid:
            raise Exception("invalid signature")
        if not token.isExpired():
            raise Exception("token expired")
        member.canLogin(self.request)
        self.credentials = UberDict(kind="member", instance=member, pk=member.pk, uuid=member.username)
        self.on_authenticated()
 
    def on_authenticated(self):
        if self.debug:
            logger.info(F"authenticated {self.credentials.kind}: {self.credentials.uuid}")
        if self.credentials.kind == "member":
            channel_key = self.redis.subscribe("user", self.facility, self.credentials.uuid)
            self.forwardPending(channel_key)
        else:
            channel_key = self.redis.subscribe(self.credentials.kind, self.facility, self.credentials.uuid)
            self.forwardPending(channel_key)

        self.redis.publishModelOnline(self.credentials.kind, self.credentials.pk, only_one=self.credentials.only_one)
        self.refreshFDs()
        if self.credentials and self.credentials.instance:
            if hasattr(self.credentials.instance, "on_ws_online"):
                self.credentials.instance.on_ws_online()

    def forwardPending(self, channel_key):
        msg = self.redis.getPendingMessage(channel_key)
        if msg:
            if self.debug:
                logger.info("pending messages", repr(msg))
            channel, pk = self.parseChannel(channel_key)
            self.sendToWS(channel, msg, pk)

    def on_resubscribe(self, msg):
        for ch in msg.channels:
            self.redis.subscribe(ch.channel, self.facility, ch.pk)

    def getAppModel(self, app_model):
        if app_model not in MODEL_CACHE:
            app_label, model_name = app_model.split('.')
            MODEL_CACHE[app_model] = apps.get_model(app_label, model_name)
        return MODEL_CACHE[app_model]

    def getAuthenticator(self, channel):
        auther_path = private_settings.WS4REDIS_AUTHENTICATORS.get(channel, None)
        if auther_path is not None:
            return self.getAppModel(auther_path)
        return None

    def on_subscribe(self, msg):
        pks = self.canSubscribeTo(msg)
        if bool(pks):
            for pk in pks:
                channel_key = self.redis.subscribe(msg.channel, self.facility, pk)
        else:
            logger.warning("subscribe permission denied", msg)
            self.sendToWS(msg.channel, dict(error="subscribe permission denied"))

    def canSubscribeTo(self, msg):
        ch_model = private_settings.WS4REDIS_CHANNELS.get(msg.channel, None)
        if not ch_model:
            return None
        if ch_model == "any":
            # this allows anything to be sent
            return [msg.pk]
        Model = self.getAppModel(ch_model)
        if not hasattr(Model, "canSubscribeTo"):
            return None
        return Model.canSubscribeTo(self.credentials, msg)

    def on_unsubscribe(self, msg):
        self.redis.unsubscribe(msg.channel, self.facility, msg.pk)

    def on_publish(self, msg):
        if self.canPublishTo(msg):
            self.redis.publish(msg.message, channel=msg.channel, pk=msg.pk)
        else:
            logger.warning("publish permission denied", msg)
            self.sendToWS(msg.channel, dict(error="publish permission denied"))

    def canPublishTo(self, msg):
        ch_model = private_settings.WS4REDIS_CHANNELS.get(msg.channel, None)
        if not ch_model:
            # you can publish to any channels you may create
            return True
        Model = self.getAppModel(ch_model)
        if not hasattr(Model, "canPublishTo"):
            return False
        return Model.canPublishTo(self.credentials, msg)

    def on_ws_msg(self, raw_data):
        if isinstance(raw_data, bytes):
            raw_data = util.toString(raw_data)
        if raw_data == self._heart_beat:
            # echo the heartbeat
            self.websocket.send(self._heart_beat)
            return
        if self.debug:
            logger.info("on_ws_msg", repr(raw_data))
        dmsg = UberDict.fromJSON(raw_data, ignore_errors=True)
        if dmsg.action:
            # this is a special message that we want to handle directly
            if dmsg.action == "auth":
                self.on_auth(dmsg)
            elif dmsg.action == "subscribe":
                self.on_subscribe(dmsg)
            elif dmsg.action == "unsubscribe":
                self.on_unsubscribe(dmsg)
            elif dmsg.action == "resubscribe":
                self.on_resubscribe(dmsg)
            elif dmsg.action == "publish":
                self.on_publish(dmsg)
            elif dmsg.action == "save":
                self.on_save(dmsg)
            elif dmsg.channel:
                self.on_channel_msg(dmsg)

    def on_save(self, msg):
        if self.credentials and self.credentials.instance:
            if hasattr(self.credentials.instance, "on_ws_save"):
                resp = UberDict(id=msg.id, name="save") 
                if msg.echo:
                    resp.data = msg.data
                resp.status = self.credentials.instance.on_ws_save(msg)
                self.sendToWS(self.credentials.kind, resp)

    def on_channel_msg(self, msg):
        if not self.canPublishTo(msg):
            logger.warning("on_channel_msg permission denied", msg)
            self.sendToWS(msg.channel, dict(error="cannot publish to channel"))
            return None
        ch_model = private_settings.WS4REDIS_CHANNELS.get(msg.channel, None)
        if ch_model is None:
            return None
        Model = self.getAppModel(ch_model)
        if not hasattr(Model, "onWS4RedisMessage"):
            logger.warning(f"{msg.channel} does not support onWS4RedisMessage")
            self.sendToWS(msg.channel, dict(error="channel does not support onWS4RedisMessage"))
            return None
        Model.onWS4RedisMessage(self.credentials, msg, self)

    def on_redis_pending(self):
        sub_resp = self.redis.getSubMessage()
        if sub_resp:
            if self.debug:
                logger.info("incoming redis msg", sub_resp)
            self.on_redis_msg(sub_resp)

    def sendToWS(self, channel, message, pk=None):
        msg = UberDict(channel=channel)
        if pk is not None:
            msg.pk = pk
        if isinstance(message, (str, bytes)):
            msg.message = UberDict.fromJSON(message, ignore_errors=True)
        if not msg.message:
            msg.message = message
        elif msg.message.name == "logout" and msg.message.pk == self.credentials.pk:
            raise Exception("websocket is being logged out")
        self.websocket.send(msg.toJSON(as_string=True)) 

    def on_redis_msg(self, sub_resp):
        if isinstance(sub_resp, list):
            if sub_resp[0] == b'subscribe':
                # this is succesfull subscriptions
                # notify ws
                channel, pk = self.parseChannel(sub_resp[1].decode())
                msg = UberDict(name="subscribed", channel=channel, status=sub_resp[2])
                if pk is not None:
                    msg.pk = pk
                self.websocket.send(msg.toJSON(as_string=True))
                return
            elif sub_resp[0] == b'unsubscribe':
                # this is succesfull subscriptions
                # notify ws
                channel, pk = self.parseChannel(sub_resp[1].decode())
                msg = UberDict(name="unsubscribed", channel=channel, status=sub_resp[2])
                if pk is not None:
                    msg.pk = pk
                self.websocket.send(msg.toJSON(as_string=True))
                return
            elif sub_resp[0] == b'message':
                channel, pk = self.parseChannel(sub_resp[1].decode())
                self.sendToWS(channel, sub_resp[2].decode(), pk)
                return
        sendmsg = RedisMessage(sub_resp)
        if self.debug:
            logger.info(sub_resp)
        if sendmsg:
            if self.debug:
                logger.info("pushing to websocket", sendmsg)
            self.websocket.send(sendmsg)

    def parseChannel(self, channel):
        fields = channel.split(":")
        fields.pop(0)
        fields.pop()
        if len(fields) == 1:
            return fields[0], None
        return fields[0], fields[1]

    def on_ws_pending(self):
        try:
            self.last_msg = self.websocket.receive()
            if bool(self.last_msg):
                self.on_ws_msg(self.last_msg)
        except Exception:
            logger.exception()
            logger.error("unable to recv on ws... flushing")
            self.websocket.flush()

    def checkHeartbeat(self):
        beat_delta = time.time() - self.last_beat
        if beat_delta > 30.0:
            self.last_beat = time.time()
            if private_settings.WS4REDIS_HEARTBEAT and not self.websocket.closed:
                # logger.info("send heartbeat")
                self.websocket.send(private_settings.WS4REDIS_HEARTBEAT)

    def release(self):
        self.redis.release()
        self.listening_fds = None
        if self.websocket:
            self.websocket.close(code=1001, message='Websocket Closed')
        if self.credentials and self.credentials.instance:
            if hasattr(self.credentials.instance, "on_ws_offline"):
                self.credentials.instance.on_ws_offline()

    def handleComs(self):
        # check if token is in url
        if not ALLOW_ANY_FACILITY and self.facility not in private_settings.WS4REDIS_FACILITIES:
            self.sendToWS("facility", dict(error=f"facility name not allowed: '{self.facility}'.  Please check your connection URL."))
            return

        websocket_fd = self.websocket.get_file_descriptor()
        self.listening_fds = [websocket_fd]
        if callable(private_settings.URL_AUTHENTICATOR):
            private_settings.URL_AUTHENTICATOR(self)
        no_cred_count = 0
        
        while self.websocket and not self.websocket.closed:
            ready = self.server.select(self.listening_fds, [], [], 4.0)[0]
            if not ready:
                self.websocket.flush()
            if not self.credentials:
                no_cred_count += 1
                if no_cred_count > 10:
                    logger.error(f"{self.ip} has sent no credentials", self.ua)
                    no_cred_count = 0
                    self.sendToWS("member", dict(error="no credentials received"))
                    
            for fd in ready:
                if fd == websocket_fd:
                    self.on_ws_pending()
                else:
                    self.on_redis_pending()



"""
It is critical to store everything in UTC.
Timezone that have DST (Daylight Savings)
or repeat an hour once a year.

When calculating actual elapsed time if using timezone aware objects
convert to epoch or use

"""


from dateutil.parser import parse as date_parser
from datetime import date, datetime, timedelta
import calendar
import pytz
import time
import re


def tz_datetime(timezone_str, year=2023, month=1, day=1, hour=0, minute=0, second=0):
    timezone = pytz.timezone(timezone_str)
    return datetime(
        year=year, month=month, day=day, 
        hour=hour, minute=minute, second=second,
        tzinfo=timezone)


def delta_seconds(d1, d2):
    if isinstance(d1, (int, float)):
        return d1 - d2
    if d1.tzinfo is not None or d2.tzinfo is not None:
        if d1.tzinfo is None:
            d1 = d2.tzinfo.localize(d1)
        elif d2.tzinfo is None:
            d2 = d1.tzinfo.localize(d2)
    return (d1 - d2).total_seconds()


def delta_hours(start, end):
    return int(delta_seconds(start, end) / 60 / 60)


def delta_days(start, end):
    return int(delta_seconds(start, end) / 60 / 60 / 24)


def epoch_offset(epoch_time, minutes=None, hours=None, days=None, timezone_str='UTC'):
    # Convert epoch time to timezone-aware datetime
    dt = datetime.fromtimestamp(epoch_time)
    # Calculate the same time on the previous day
    new_dt = offsetDate(dt, minutes=minutes, hours=hours, days=days)
    return new_dt.timestamp()


def offsetDate(when=None, minutes=None, hours=None, days=None):
    """
    Generates a new datetime object based on the offesets provides
    offsets can be negative or positive
    """
    params = {}
    if minutes is not None:
        params["minutes"] = minutes
    if hours is not None:
        params["hours"] = hours
    if days is not None:
        params["days"] = days
    if when is None:
        when = datetime.now()
    return when + timedelta(**params)


def convertToEpoch(dt):
    return time.mktime(dt.timetuple())


def getShortTZ(zone, when=None):
    timezone = pytz.timezone(zone)
    if not when:
        when = datetime.today()
    ltz = timezone.localize(when, is_dst=None)
    return ltz.tzname()


def convertToLocalTime(zone, when=None, tz_aware=False):
    """
    convert timezones in python can result in minutes shifting...
    just do a simple hour offset with timedelta
    """
    if when is None:
        when = datetime.today()

    offset = getTimeZoneOffset(zone, when)
    if offset >= 0:
        when = when - timedelta(hours=offset)
    else:
        when = when + timedelta(hours=offset)
    if tz_aware:
        return when.replace(tzinfo=pytz.timezone(zone))
    return when


def convertToUTC(zone, when=None):
    # this works because it gets around the issue with naive vs aware
    offset = getTimeZoneOffset(zone, when)
    if offset >= 0:
        when = when + timedelta(hours=offset)
    else:
        when = when - timedelta(hours=offset)
    return when


def convertToUTCEx(zone, when):
    local = pytz.timezone(zone)
    local_dt = local.localize(when, is_dst=None)
    return local_dt.astimezone(pytz.utc)


def getUTC(zone, when=None):
    timezone = pytz.timezone(zone)
    if when is None:
        when = datetime.today()
    return timezone.utcoffset(when)


def getTimeZoneOffset(zone, when=None, hour=None, dst=True):
    if zone is None:
        zone = "UTC"
    timezone = pytz.timezone(zone)
    if not when:
        when = datetime.today()
    timestamp = when
    if hour != None:
        when = when.replace(tzinfo=pytz.UTC, hour=hour)
    else:
        hour = 0
        when = when.replace(tzinfo=pytz.UTC)

    offset = int(when.astimezone(timezone).utcoffset().total_seconds()/3600)
    if not dst:
        offset = when.astimezone(timezone).utcoffset() - timezone.dst(timestamp)
        offset = int(offset.total_seconds()/3600)
    # if hour != None:
    offset = abs(offset) + hour
    if offset >= 24:
        offset -= 24
    return offset


def diffNow(dt):
    return diffSeconds(datetime.now(), dt)


def diffSeconds(t1, t2):
    return (t1 - t2).total_seconds()


def diffMinutes(t1, t2):
    diff = t1 - t2
    days, seconds = diff.days, diff.seconds
    hours = (days * 24)
    return (seconds / 60) + (hours * 60)


def diffHours(t1, t2):
    return diffMinutes(t1, t2) / 60


def next_weekday(d, weekday):
    return d + timedelta(days=((weekday - d.weekday()) + 7) % 7)


def prev_weekday(d, weekday):
    return d - timedelta(days=((d.weekday() - weekday) + 7) % 7)


def getWeek(start, start_day=0):
    # TODO allow selection of start day of week, where Monday is 0 and Sunday is 6
    # get records between last Monday and next Sunday
    week_start = start + timedelta(days=-start.weekday())
    week_end = start + timedelta(days=-start.weekday() - 1, weeks=1)
    return week_start, week_end


def getStartOfMonth(d):
    return d.replace(day=1)


def nextMonth(d):
    return getStartOfMonth(getStartOfMonth(d) + timedelta(days=32))


def getEndOfMonth(start):
    start = start.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    end = start + timedelta(days=calendar.monthrange(start.year, start.month)[1])
    end = end.replace(hour=0, minute=0, second=0, microsecond=0)
    return end


def parseDate(date_str, is_future=False, is_past=False, month_end=True, as_date=False):
    res = parseDateTime(date_str, is_future, is_past, month_end)
    if as_date and res:
        return date(res.year, res.month, res.day)
    return res


def parseDateTime(date_str, is_future=False, is_past=False, month_end=True):
    if type(date_str) in [str, str]:
        dt = None
        if len(date_str) > 6 and date_str.count('.') <= 1 and date_str.split('.')[0].isdigit():
            try:
                f = float(date_str)
                return datetime.utcfromtimestamp(f)
            except Exception as err:
                print(err)
        else:
            fix_month = False
            if date_str.count('-') or date_str.count('/') or date_str.count(' ') or date_str.count('.'):
                dts = re.split('/|-| |\.',date_str)
                if len(dts) == 2:
                    date_str = "{0}-{1}-01".format(dts[0], dts[1])
                else:
                    "-".join(dts)
            elif len(date_str) == 4:
                fix_month = month_end
                date_str = date_str[:2] + "-01-" + date_str[2:4]
            elif len(date_str) == 3:
                date_str = date_str[0] + "/1/" + date_str[-2:]
            elif len(date_str) == 6:
                date_str = date_str[:2] + "-" + date_str[2:4] +  "-" + date_str[4:6]
        try:
            # print "parse date: {}".format(date_str)
            dt = date_parser(date_str)
            if is_future and dt.year < 2000:
                dt = dt.replace(year=dt.year+100)
            elif is_past and dt.year >= 2000:
                dt = dt.replace(year=dt.year-100)
            # print dt
            if fix_month:
                dt = dt.replace(day=calendar.monthrange(dt.year, dt.month)[1])
            # print dt
            return dt
        except BaseException:
            pass

    elif type(date_str) in [date, datetime]:
        return date_str
    elif type(date_str) in [float, int]:
        return datetime.utcfromtimestamp(date_str)
    return None


def getDateRangeZ(start, end=None, kind=None, zone=None, hour=0, eod=None, end_eod=None):
    return getDateRange(start, end, kind, zone, hour, eod, end_eod)


def getDateRange(start, end=None, kind=None, zone=None, hour=0, eod=None, end_eod=None):
    if start is None:
        start = datetime.now()
    if zone is None or zone == "":
        zone = "UTC"
    if start == end:
        end = None
    if eod is not None:
        hour = eod

    start = parseDate(start)
    if end:
        end = parseDate(end)
        if start == end:
            end = None

    if kind and kind != "second":
        start = start.replace(minute=0, second=0, microsecond=0)
        if kind == "hour":
            end = start + timedelta(hours=1)
        elif kind == "day":
            start = start.replace(hour=0)
            end = start + timedelta(days=1)
        elif kind == "week":
            start = start.replace(hour=0)
            start, end = getWeek(start)
        elif kind == "month":
            start = start.replace(hour=0, day=1)
            end = getEndOfMonth(start)
        elif kind == "year":
            start = start.replace(hour=0, day=1, month=1)
            end = getEndOfMonth(start.replace(month=12))
        elif type(kind) is int or (isinstance(kind, str) and kind.isdigit()):
            end = start + timedelta(days=1)
            start = end - timedelta(days=int(kind))
    
    if end is None:
        end = start + timedelta(hours=24)

    if zone and zone.lower() == "utc":
        zone = None
        if not kind and eod:
            hour = None
    # now lets convert our times to the zone
    if zone or hour:
        if zone is None:
            zone = "UTC"
        offset = getTimeZoneOffset(zone, start, hour=hour)
        if offset:
            start = start + timedelta(hours=offset)
        if end_eod:
            hour = end_eod
        offset = getTimeZoneOffset(zone, end, hour=hour)
        if offset:
            end = end + timedelta(hours=offset)
    return start, end
